/*
 * plcd daemon
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <sched.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "macros.h"
#include "xmalloc.h"
#include "buffer.h"
#include "list.h"
#include "tree.h"
#include "str.h"
#include "sock-unix.h"
#include "prioq.h"
#include "io.h"
#include "timeout.h"
#include "par.h"
#include "pimage.h"
#include "log.h"
#include "signal-event.h"
#include "pclt.h"
#include "pi-proto.h"
#include "misc.h"

#include "main.h"
#include "connector.h"
#include "conf.h"

#define DEV_NULL                "/dev/null"

/* Default limits */
#define MEGABYTE                (1024 * 1024)
#define MAX_MEMLOCK             (10 * MEGABYTE)	/* 10 megabytes seems reasonable */
#define MAX_RTPRIO              99		/* maximum realtime performance */
#define MAX_NOFILE              1024		/* seems reasonable limit */
#define MAX_RTTIME              60		/* one minute in real-time is more than enough */

#define STACK_BYTES             8192
#define MEMORY_PREALLOC         (4 * MEGABYTE)

#define	timeval_to_usec(tv)     ((tv).tv_sec * 1000000 + (tv).tv_usec)

struct limit_options {
	int		resource;
	char		*resource_name;
	struct		rlimit limit;
};

/* limit options  */
struct limit_options limit_options[] = {
	{ RLIMIT_MEMLOCK, "RLIMIT_MEMLOCK", { MAX_MEMLOCK, MAX_MEMLOCK } },
	{ RLIMIT_NOFILE, "RLIMIT_NOFILE", { MAX_NOFILE, MAX_NOFILE } },
#ifdef RLIMIT_RTPRIO
	{ RLIMIT_RTPRIO, "RLIMIT_RTPRIO", {MAX_RTPRIO, MAX_RTPRIO}},
#endif
#ifdef RLIMIT_RTTIME
	{ RLIMIT_RTTIME, "RLIMIT_RTTIME", {MAX_RTTIME, MAX_RTTIME}},
#endif
};

/* performance stats */
struct performance_counters {
	unsigned long		io_read_usec;
	unsigned long		io_write_usec;
	unsigned long		cycle_usec;
	unsigned long		total_usec_last;
	unsigned long		total_usec_max;
	unsigned long		total_usec_min;
	unsigned long		cycles_run;
	struct par_struct	*par_cycle_msec;
	struct par_struct	*par_read_msec;
	struct par_struct	*par_write_msec;
	struct par_struct	*par_min_msec;
	struct par_struct	*par_max_msec;
	struct par_struct	*par_last_msec;
};

static int terminate = 0;
static int realtime_scheduler_enabled = 0;
static int realtime_scheduler_error = 0;
static int retry_tout_msec = 0;
static int cyclemsec = 0;
static struct timeout *retry_tout;
static struct io_context *io_wait;
static struct performance_counters io_stats;

extern slist_t pclt_options_list;

static void signal_handler(int signo, void *opaque)
{
	switch (signo) {
	case SIGALRM:
		/* When SIGALRM is received we
		 * must stop waiting and repeat
		 * main cycle.
		 */
		io_loop_stop(io_wait);
		break;
	case SIGTERM:
	case SIGINT:
		info("exiting by signal %d", signo);
		io_loop_stop(io_wait);
		terminate = TRUE;
		break;
	default:
		info("unhandled signal %d", signo);
		break;
	}
}

static void event_signal(int fd, short events, void *opaque)
{
	if (events & EVENT_ERROR)
		error("signal fd=%d reports error", fd);
	if (events & EVENT_READ)
		sigevent_run();
}

static void fail_pages_in_stack()
{
	int res;
	struct rlimit rlim;
	size_t i, max, pagesize;
	unsigned char buffer[STACK_BYTES];

	res = getrlimit(RLIMIT_STACK, &rlim);
	if (res) {
		warn("getrlimit() RLIMIT_STACK: %s", strerror(errno));
		return;
	}
	/* do not overflow limit */
	max = rlim.rlim_cur > STACK_BYTES ? STACK_BYTES : rlim.rlim_cur;
	/* get pagesize */
	pagesize = getpagesize();
	/* touch every page first byte in the stack */
	for (i = 0; i < max; i += pagesize)
		buffer[i] = buffer[i];

	verbose("MEMORY: prefault %ld pages (%ld bytes) in stack", (unsigned long) max / pagesize + (max % pagesize ? 1: 0), (unsigned long) max);
}

static void ensure_memory_for_malloc(size_t size)
{
	int i, res;
	unsigned char *buffer;
	size_t pagesize;
	struct rusage rusage;

	res = mlockall(MCL_CURRENT | MCL_FUTURE);
	if (res)
		warn("mlockall(): %s", strerror(errno));
	/* Disable malloc trimming */
	mallopt(M_TRIM_THRESHOLD, -1);
	/* Disable malloc using mmap */
	mallopt(M_MMAP_MAX, 0);
	/* allocate buffer and touch every first byte of every page */
	buffer = malloc(size);
	if (buffer) {
		pagesize = getpagesize();
		for (i = 0; i < size; i += pagesize)
			buffer[i] = 0;
		free(buffer);
	} else {
		warn("malloc() failed to ensure %ld bytes: %s", (unsigned long) size, strerror(errno));
	}
	getrusage(RUSAGE_SELF, &rusage);
	verbose("MEMORY: %ld bytes locked major faults: %ld minor faults: %ld", (unsigned long) size, rusage.ru_minflt, rusage.ru_majflt);
}

static void drop_privileges(uid_t user, gid_t group)
{
	int res;

	if (getuid() != 0)
		return;

	if (group > 0) {
		debug("dropping to gid=%d", group);
		res = setgid(group);
		if (res)
			warn("setgid(): %s", strerror(errno));
	}

	if (user > 0) {
		debug("dropping to uid=%d", user);
		res = setuid(user);
		if (res)
			warn("setuid(): %s", strerror(errno));
	}
}

void periodic_timer_enable(int msec)
{
	int err;
	struct itimerval itimer;

	itimer.it_value.tv_sec = msec / 1000;
	itimer.it_value.tv_usec = (msec % 1000) * 1000;
	itimer.it_interval.tv_sec = msec / 1000;
	itimer.it_interval.tv_usec = (msec % 1000) * 1000;
	err = setitimer(ITIMER_REAL, &itimer, NULL);
	if (err)
		warn("periodic_timer: setitimer() failed: %s", strerror(errno));
}

void periodic_timer_disable()
{
	int err;
	struct itimerval itimer;

	itimer.it_value.tv_sec = 0;
	itimer.it_value.tv_usec = 0;
	itimer.it_interval.tv_sec = 0;
	itimer.it_interval.tv_usec = 0;
	err = setitimer(ITIMER_REAL, &itimer, NULL);
	if (err)
		warn("periodic_timer: setitimer() failed: %s", strerror(errno));
}

void realtime_enter(int priority)
{
	int res;
	struct sched_param parameters;

	if (realtime_scheduler_error)
		return;

	if (realtime_scheduler_enabled) {
		warn("realtime_enter: I am already in SCHED_RR.");
		return;
	}

	parameters.sched_priority = priority;
	res = sched_setscheduler(0, SCHED_RR, &parameters);
	if (res) {
		realtime_scheduler_error = TRUE;
		warn("realtime_enter: Failed to enter RT SCHED_RR scheduler: %s", strerror(errno));
		return;
	}

	realtime_scheduler_enabled = TRUE;
	info("RT scheduler priority %d", priority);
}

void realtime_exit()
{
	int res;
	struct sched_param parameters;

	if (realtime_scheduler_error)
		return;

	if (!realtime_scheduler_enabled) {
		warn("realtime_exit: I am not in SCHED_RR");
		return;
	}

	parameters.sched_priority = 0;
	res = sched_setscheduler(0, SCHED_OTHER, &parameters);
	if (res)
		warn("realtime_exit: Failed to enter SCHED_OTHER back: %s", strerror(errno));

	realtime_scheduler_enabled = FALSE;
	info("exited RT scheduler");
}

static void init_connector(char *connectorfile)
{
	int res;
	struct stat st;

	if (stat(connectorfile, &st) == 0 && S_ISREG(st.st_mode)) {
		res = connector_init(connectorfile);
		if (res) {
			warn("Failed to init connector application. I will not run connector. Please check configuration file '%s'.", connectorfile);
			return;
		}
	} else {
		verbose("Connector application will not run because no file '%s' exists.", connectorfile);
		return;
	}

	info("Initialized connector using configuration file '%s'.", connectorfile);
}

#define create_sys_param(par, name, type) \
{ \
  if (par_parameter_add(&(par), "/sys", name, type) != 0) \
    error("can't create system parameter '/sys/%s'", name); \
}

/* creates system statistics parameters */
static void init_stat_params()
{
	create_sys_param(io_stats.par_read_msec, "cycle_read_msec", PAR_INT);
	create_sys_param(io_stats.par_write_msec, "cycle_write_msec", PAR_INT);
	create_sys_param(io_stats.par_cycle_msec, "cycle_msec", PAR_INT);
	create_sys_param(io_stats.par_min_msec, "cycle_min_msec", PAR_INT);
	create_sys_param(io_stats.par_max_msec, "cycle_max_msec", PAR_INT);
	create_sys_param(io_stats.par_last_msec, "cycle_last_msec", PAR_INT);
}

/* update system statistics parameters */
static void update_stat_params()
{
	par_write(io_stats.par_read_msec, (par_value_t)((int) (io_stats.io_read_usec / 1000)));
	par_write(io_stats.par_write_msec, (par_value_t)((int) (io_stats.io_write_usec / 1000)));
	par_write(io_stats.par_cycle_msec, (par_value_t)((int) (io_stats.cycle_usec / 1000)));
	par_write(io_stats.par_max_msec, (par_value_t)((int) (io_stats.total_usec_max / 1000)));
	par_write(io_stats.par_min_msec, (par_value_t)((int) (io_stats.total_usec_min / 1000)));
	par_write(io_stats.par_last_msec, (par_value_t)((int) (io_stats.total_usec_last / 1000)));
}

static void apply_limit_options()
{
	int i, err;

	for (i = 0; i < N_ELEMS(limit_options); i++) {
		err = setrlimit(limit_options[i].resource, &limit_options[i].limit);
		if (err)
			warn("setrlimit() %s: %s", limit_options[i].resource_name, strerror(errno));
		else
			info("setrlimi() %s hard=%ld soft=%ld", limit_options[i].resource_name, limit_options[i].limit.rlim_max, limit_options[i].limit.rlim_cur);
	}
}

static int list_client_callback(const char *sockfile, int state, int status)
{
	char *str = "unknown";

	if (state == CLIENT_DISCONNECTED) {
		switch (status) {
		case STATUS_OK: str = "disconnected"; break;
		case STATUS_TIMEOUT: str = "response timeout"; break;
		case STATUS_BAD_COMMAND: str = "bad command"; break;
		case STATUS_ERANGE: str = "read/write out of range"; break;
		case STATUS_MALFORMED: str = "malformed response/request"; break;
		case STATUS_NOT_CONNECTED: str = "not connected"; break;
		case STATUS_NOT_AVAIL: str = "data not available"; break;
		case STATUS_SIDE_MISMATCH: str = "master/slave mismatch"; break;
		}
		debug("PI socket '%s' disconnected: cause=%s", sockfile, str);
	}
	return 0;
}

static void retry_timeout_handler(int fd, short event, void *opaque)
{
	if (event & EVENT_TIMEOUT) {
		/* list sessions */
		pclt_enumerate(list_client_callback);
		/* try to connect to disabled PI server sessions */
		pclt_try_reconnect();
		/* re-enable timer */
		if (retry_tout_msec > 0)
			timeout_add(retry_tout, retry_tout_msec);
	}
}

static void init_pservers( )
{
	int res;
	slist_t tmp_list, list;
	struct par_struct *par;
	str_t parent;
	str_t container;
	str_t container_in;
	str_t container_out;
	struct pclt *pclt;

	parent = str_new();
	container_in = str_new();
	container_out = str_new();

	for (list = pclt_options_list; list != NULL; list = list->next) {
		struct pclt_options *opt = (struct pclt_options *) list->data;
		pclt = pclt_new(opt->socketfile, opt->timeout, opt->inputs, opt->outputs, opt->mode, opt->side);
		if (pclt != NULL)
			info("Connected to PI server '%s' sock '%s'.", opt->name, opt->socketfile);
		else {
			warn("Failed to connect to PI server '%s' sock '%s'.", opt->name, opt->socketfile);
			continue;
		}
		par_container_add("/pi", opt->name);
		str_reset(parent);
		str_reset(container_in);
		str_reset(container_out);
		str_snprintf(parent, "/pi/%s", opt->name);
		str_append_str(container_in, parent);
		str_append_str(container_out, parent);
		str_append(container_in, "/in");
		str_append(container_out, "/out");
		par_container_add(str_ptr(parent), "in");
		par_container_add(str_ptr(parent), "out");
		for (tmp_list = opt->par_options_list; tmp_list != NULL; tmp_list = tmp_list->next) {
			struct par_options *par_opt = (struct par_options *) tmp_list->data;
			if (par_opt->dir == PAR_DIR_IN)
				container = container_in;
			else if (par_opt->dir == PAR_DIR_OUT)
				container = container_out;
			else {
				warn("Unknown parameter direction: '%s' '%s'.", opt->name, par_opt->name);
				continue;
			}
			res = par_parameter_add(&par, str_ptr(container), par_opt->name, par_opt->type);
			if (!res) {
				/* write initial parameter value */
				par_write(par, par_opt->initial_value);
				verbose("Created parameter '%s/%s' %s.", str_ptr(container), par_opt->name, par_type2str( par_opt->type));
				res = pclt_add_par(pclt, par, par_opt->dir, par_opt->offset, par_opt->bit);
				if (res)
					warn("Failed to link parameter '%s/%s' with PI client.", str_ptr(container), par_opt->name);
			} else {
				warn("Failed to create parameter /%s/%s.", str_ptr(container), par_opt->name);
			}
		}
	}
	str_free(&container_in);
	str_free(&container_out);
	str_free(&parent);
}

void print_usage()
{
	printf("Usage: plcd [-vsihu] [-u user] [-g group] [-p prio] [-c file] \n"
	       "            [-f file] [-l loglevel] [-t period] [-r retry] \n"
	       " \n"
	       "  -v                    print version \n"
	       "  -s                    use syslog \n"
	       "  -i                    stay interactive do not become daemon \n"
	       "  -h                    print this help \n"
	       "  -u user               user name or uid to drop privileges to \n"
	       "  -g group              group name or gid to drop privileges to \n"
	       "  -p prio               RT scheduler priority in range %d..%d \n"
	       "  -c file               path to plcd.conf \n"
	       "  -f file               path to connector.conf \n"
	       "  -l loglevel           log level [debug|verbose|info|warn|error] \n"
	       "  -t period             minimum cycle time period msec [%d..%d] \n"
	       "  -r retry              reconnect disabled PI sessions in seconds %d..%d or 0 to disable \n"
	       " \n",
		sched_get_priority_min(SCHED_RR), sched_get_priority_max(SCHED_RR),
		MIN_CYCLE_MSEC, MAX_CYCLE_MSEC, MIN_RETRY_SECS, MAX_RETRY_SECS);
}

static void run_cycle_once()
{
	struct timeval tv_res, tv_prev, tv_now;
	struct timeval tv_begin, tv_end;
	unsigned long total_usec;

	/* Main cycle routines are executed here. */
	/* We acquire performance statistics and save it in /sys. */

	gettimeofday(&tv_prev, NULL);
	tv_begin = tv_prev;

	/* notify slave servers we want their inputs */
	pclt_prepare_inputs(PI_SIDE_SLAVE);

	/* read inputs from slaves and update image points */
	pclt_read_inputs();

	/* run connector and copy slave inputs to master inputs (if any) */
	connector_run(CONNECTOR_BEGIN);

	/* notify master servers we will put inputs */
	pclt_prepare_inputs(PI_SIDE_MASTER);

	/* write inputs to master servers */
	pclt_write_inputs();

	/* notify master servers we want to read their outputs */
	pclt_prepare_outputs(PI_SIDE_MASTER);

	/* read master servers outputs */
	pclt_read_outputs();

	gettimeofday(&tv_now, NULL);
	timersub(&tv_now, &tv_prev, &tv_res);
	io_stats.io_read_usec = timeval_to_usec(tv_res);
	tv_prev = tv_now;

	/* ----- Application here ----- */

	/* ---------------------------- */

	gettimeofday(&tv_now, NULL);
	timersub(&tv_now, &tv_prev, &tv_res);
	io_stats.cycle_usec = timeval_to_usec(tv_res);
	tv_prev = tv_now;

	/* run connector and copy master outputs to slave outputs */
	connector_run(CONNECTOR_END);

	/* prepare to write slave outputs */
	pclt_prepare_outputs(PI_SIDE_SLAVE);

	/* write outputs to slaves */
	pclt_write_outputs();

	gettimeofday(&tv_now, NULL);
	timersub(&tv_now, &tv_prev, &tv_res);
	io_stats.io_write_usec = timeval_to_usec(tv_res);

	/* wait for cycle to end */
	if (cyclemsec)
		io_loop_run(io_wait);
	else
		sigevent_run();

	gettimeofday(&tv_end, NULL);
	timersub(&tv_end, &tv_begin, &tv_res);
	total_usec = timeval_to_usec(tv_res);
	if (io_stats.total_usec_min) {
		if (total_usec < io_stats.total_usec_min)
			io_stats.total_usec_min = total_usec;
	} else
		io_stats.total_usec_min = total_usec;
	if (io_stats.total_usec_max) {
		if (total_usec > io_stats.total_usec_max)
			io_stats.total_usec_max = total_usec;
	} else
		io_stats.total_usec_max = total_usec;
	io_stats.total_usec_last = total_usec;
	io_stats.cycles_run++;
	/* update /sys performance statistics  */
	update_stat_params();
}

#define HAVE_OPTION_UID		0x0001
#define HAVE_OPTION_GID		0x0002
#define HAVE_OPTION_LOGLEVEL	0x0004
#define HAVE_OPTION_CYCLEMSEC	0x0008
#define HAVE_OPTION_PRIORITY	0x0010
#define HAVE_OPTION_CONFIGFILE	0x0020
#define HAVE_OPTION_CONNECTOR	0x0040
#define HAVE_OPTION_RETRYMSEC	0x0080
#define HAVE_OPTION_FACILITY	0x0100

int main(int argc, char *argv[])
{
	int c, res;
	char *cp;
	int log_to_syslog;
	int no_detach;
	uid_t uid;
	gid_t gid;
	int priority;
	log_level_t loglevel;
	log_facility_t facility;
	char configfile[256];
	char connectorfile[256];
	int retry_timeout_msec;
	struct passwd pwd;
	struct group grp;
	char filepath[256];
	int option_flags = 0;
	struct options options;

	/* Init defaults. */
	uid = gid = 0;
	log_to_syslog = FALSE;
	no_detach = FALSE;
	cyclemsec = DEFAULT_PERIOD;
	loglevel = LOG_LEVEL_ERROR;
	priority = DEFAULT_PRIORITY;
	snprintf(configfile, sizeof(configfile), "%s/%s/%s", CONFIG_DIR, DAEMON_DIR, CONFIG_FILE);
	snprintf(connectorfile, sizeof(connectorfile), "%s/%s/%s", CONFIG_DIR, DAEMON_DIR, CONNECTOR_FILE);
	retry_timeout_msec = DEFAULT_RETRY_TIMER;
	facility = LOG_FACILITY_DAEMON;

	/* ensure stdio fds */
	sanitize_stdio();

	while ((c = getopt(argc, argv, "vshiu:g:p:c:f:l:t:")) > 0) {
		switch (c) {
		case 'v':
			printf("plcd 0.1 (Beta) revision 0\n"
				"Copyright (c) 2009, Grigoriy Sitkarev\n"
				"License GPLv2+: GNU Public License 2 or later\n"
				"This is a free software you can change and redistribute it.\n"
				"There is no warranty on any purpose of this sotware.\n"
				"Please, send bugs and suggestions to <sitkarev@unixkomi.ru>\n");
			exit(0);
			break;
		case 'i':
			no_detach = TRUE;
			break;
		case 's':
			log_to_syslog = TRUE;
			break;
		case 'h':
			print_usage();
			exit(0);
			break;
		case 'u':
			if (option_flags & HAVE_OPTION_UID) {
				printf("Option '-u' duplicated. Please provide only one user/uid to drop privileges to.\n");
				print_usage();
				exit(-1);
			}
			if (optarg == NULL) {
				warn("getopt() NULL optarg");
				break;
			}
			res = getpw_by_name_or_uid(optarg, &pwd);
			if (res) {
				printf("User or UID '%.100s' does not exists in user database.\n", optarg);
				print_usage();
				exit(-1);
			}
			uid = pwd.pw_uid;
			if ((option_flags & HAVE_OPTION_GID) == 0)
				gid = pwd.pw_gid;
			option_flags |= HAVE_OPTION_UID;
			break;
		case 'g':
			if (option_flags & HAVE_OPTION_GID) {
				printf("Option '-g' duplicated. Please provide only one group/gid to drop privileges to.\n");
				print_usage();
				exit(-1);
			}
			if (optarg == NULL) {
				warn("getopt() NULL optarg");
				break;
			}
			res = getgr_by_name_or_gid(optarg, &grp);
			if (res) {
				printf("Group or GID '%.100s' does not exists in group database.\n", optarg);
				print_usage();
				exit(-1);
			}
			gid = grp.gr_gid;
			option_flags |= HAVE_OPTION_GID;
			break;
		case 'c':
			if (option_flags & HAVE_OPTION_CONFIGFILE) {
				printf("Option '-c' duplicated. Please provide only one configuration file to use.\n");
				print_usage();
				exit(-1);
			}
			if (optarg) {
				if (*optarg == '/' || *optarg == '.')
					xstrncpy(filepath, optarg, sizeof(filepath));
				else
					snprintf(filepath, sizeof(filepath), "%s/%s%s", CONFIG_DIR, DAEMON_DIR, optarg);
				cp = canonize_path(filepath);
				xstrncpy(configfile, filepath, sizeof(configfile));
				xfree(cp);
				option_flags |= HAVE_OPTION_CONFIGFILE;
			} else {
				warn("getopt() NULL optarg");
			}
			break;
		case 'f':
			if (option_flags & HAVE_OPTION_CONNECTOR) {
				printf("Option '-f' duplicated. Please provide only one connector file to use.\n");
				print_usage();
				exit(-1);
			}
			if (optarg) {
				if (*optarg == '/' || *optarg == '.')
					xstrncpy(filepath, optarg, sizeof(filepath));
				else
					snprintf(filepath, sizeof(filepath), "%s/%s/%s", CONFIG_DIR, DAEMON_DIR, optarg);
				cp = canonize_path(filepath);
				xstrncpy(connectorfile, filepath, sizeof(connectorfile));
				xfree(cp);
				option_flags |= HAVE_OPTION_CONNECTOR;
			} else {
				warn("getopt() NULL optarg");
			}
			break;
		case 'l':
			if (option_flags & HAVE_OPTION_LOGLEVEL) {
				printf("Option '-l' duplicated. Please provide only one loglevel to use.\n");
				print_usage();
				exit(-1);
			}
			if (optarg == NULL) {
				warn("getopt() NULL optarg");
				break;
			}
			cp = strchr(optarg, '.');
			if (cp != NULL) {
				*cp++ = 0;
				if (strcmp(optarg, "daemon") == 0)
					facility = LOG_FACILITY_DAEMON;
				else if (strcmp(optarg, "local0") == 0)
					facility = LOG_FACILITY_LOCAL0;
				else if (strcmp(optarg, "local1") == 0)
					facility = LOG_FACILITY_LOCAL1;
				else if (strcmp(optarg, "local2") == 0)
					facility = LOG_FACILITY_LOCAL2;
				else if (strcmp(optarg, "local3") == 0)
					facility = LOG_FACILITY_LOCAL3;
				else if (strcmp(optarg, "local4") == 0)
					facility = LOG_FACILITY_LOCAL4;
				else if (strcmp(optarg, "local5") == 0)
					facility = LOG_FACILITY_LOCAL5;
				else if (strcmp(optarg, "local6") == 0)
					facility = LOG_FACILITY_LOCAL6;
				else if (strcmp(optarg, "local7") == 0)
					facility = LOG_FACILITY_LOCAL7;
				else if (strcmp(optarg, "syslog") == 0)
					facility = LOG_FACILITY_SYSLOG;
				else if (strcmp(optarg, "user") == 0)
					facility = LOG_FACILITY_USER;
				else {
					printf("Option '-l' provides unknown syslog facility '%.100s to me.'\n", optarg);
					print_usage();
					exit(-1);
				}
				log_to_syslog = TRUE;
				optarg = cp;
			}
			if (strcmp(optarg, "debug") == 0)
				loglevel = LOG_LEVEL_DEBUG;
			else if (strcmp(optarg, "verbose") == 0)
				loglevel = LOG_LEVEL_VERBOSE;
			else if (strcmp(optarg, "info") == 0)
				loglevel = LOG_LEVEL_INFO;
			else if (strcmp(optarg, "warn") == 0)
				loglevel = LOG_LEVEL_WARN;
			else if (strcmp(optarg, "error") == 0)
				loglevel = LOG_LEVEL_ERROR;
			else {
				printf("Option '-l' provides unknown log level '%.100s' to me.\n", optarg);
				print_usage();
				exit(-1);
			}
			option_flags |= HAVE_OPTION_LOGLEVEL;
			break;
		case 'p':
			if (option_flags |= HAVE_OPTION_PRIORITY) {
				printf("Option '-p' duplicated. Please provide only one RT priority.\n");
				print_usage();
				exit(-1);
			}
			if (optarg == NULL) {
				warn("getopt() NULL optarg");
				break;
			}
			if (str_isnumber(optarg)) {
				if (atoi(optarg) < sched_get_priority_min(SCHED_RR)) {
					printf("You have specified too small priority for RT scheduler.\n");
					print_usage();
					exit(-1);
				}
				if (atoi(optarg) > sched_get_priority_max(SCHED_RR)) {
					printf("You have specified too big priority for RT scheduler.\n");
					print_usage();
					exit(-1);
				}
				priority = atoi(optarg);
			} else {
				printf("Option '-p' required numeric argument.\n");
				print_usage();
				exit(-1);
			}
			option_flags |= HAVE_OPTION_PRIORITY;
			break;
		case 't':
			if (option_flags & HAVE_OPTION_CYCLEMSEC) {
				printf("Option '-t' duplicated. Please provide only one cycle period.\n");
				print_usage();
				exit(-1);
			}
			if (optarg == NULL) {
				warn("getopt NULL optarg");
				break;
			}
			if (str_isnumber(optarg)) {
				if (atoi(optarg) < MIN_CYCLE_MSEC) {
					printf("You have specified too small cycle time.\n");
					print_usage();
					exit(-1);
				}
				if (atoi(optarg) > MAX_CYCLE_MSEC) {
					printf("You have specified too long cycle time.\n");
					print_usage();
					exit(-1);
				}
				cyclemsec = atoi(optarg);
				option_flags |= HAVE_OPTION_CYCLEMSEC;
			} else {
				printf("Option '-t' requires numeric argument.\n");
				print_usage();
				exit(-1);
			}
			break;
		case 'r':
			if (option_flags & HAVE_OPTION_RETRYMSEC) {
				printf("Option '-r' duplicated. Please provide only one retry period.\n");
				print_usage();
				exit(-1);
			}
			if (optarg == NULL) {
				warn("getopt() NULL optarg");
				break;
			}
			if (str_isnumber(optarg)) {
				if (atoi(optarg) < MIN_RETRY_SECS) {
					printf("You have specified too small retry period.\n");
					print_usage();
					exit(-1);
				}
				if (atoi(optarg) > MAX_RETRY_SECS) {
					printf("You have specified too long retry period.\n");
					print_usage();
					exit(-1);
				}
				retry_timeout_msec = atoi(optarg) * 1000;
				option_flags |= HAVE_OPTION_RETRYMSEC;
			} else {
				printf("Option '-r' requires numeric argument.\n");
				print_usage();
				exit(-1);
			}
			break;
		default:
			print_usage();
			exit(-1);
			break;
		}
	}

	/* Read configuration. */
	res = read_config(configfile, &options);

	if (res)
		error("Failed to obtain configuration from file '%s'. Exiting.", configfile);

	/* Override options from configuration file by options provided as arguments. */
	if ((option_flags & HAVE_OPTION_LOGLEVEL) == 0 && options.level >= 0) {
		loglevel = options.level;
		if ((int) options.facility >= 0) {
			facility = options.facility;
			log_to_syslog = TRUE;
		}
	}
	if ((option_flags & HAVE_OPTION_UID) == 0 && (int) options.user >= 0)
		uid = options.user;
	if ((option_flags & HAVE_OPTION_GID) == 0 && (int) options.group >= 0)
		gid = options.group;
	if ((option_flags & HAVE_OPTION_CONNECTOR) == 0 && strlen(options.connectorfile))
		xstrncpy(connectorfile, options.connectorfile, sizeof(connectorfile));
	if ((option_flags & HAVE_OPTION_CYCLEMSEC) == 0 && options.cyclemsec >= 0)
		cyclemsec = options.cyclemsec;
	if ((option_flags & HAVE_OPTION_PRIORITY) == 0 && options.priority >= 0)
		priority = options.priority;
	if ((option_flags & HAVE_OPTION_RETRYMSEC) == 0 && options.retry_tout_msec >= 0)
		retry_timeout_msec = options.retry_tout_msec;

	/* initialize logger */
	log_init(argv[0], loglevel, log_to_syslog, facility, NULL);

	/* set reasonable limits */
	apply_limit_options();
	
	/* Prevent paging by locking all memory and future pages in memory. */
	fail_pages_in_stack();
	ensure_memory_for_malloc(MEMORY_PREALLOC);
	
	/* set up signals and handlers */
	sigevent_ignore(SIGPIPE);
	sigevent_ignore(SIGHUP);
	sigevent_add(SIGTERM, signal_handler, NULL, 0);
	if (no_detach)
		sigevent_add(SIGINT, signal_handler, NULL, 0);
	else
		sigevent_ignore(SIGINT);
	if (cyclemsec) {
		io_wait = io_new();
		sigevent_add(SIGALRM, signal_handler, NULL, 0);
	} else
		sigevent_ignore(SIGALRM);
	sigevent_init();
	
	/* initialize timeout queue */
	timeout_init();
	
	/* watch events on signal fd */
	if (io_wait)
		io_add(io_wait, sigevent_fd(), EVENT_READ, event_signal, NULL);

	/* Initialize parameters tree and create default containers. */
	par_tree_create();
	par_container_add("/", "pi");
	par_container_add("/", "logic");
	par_container_add("/", "sys");

	/* create statistics parameters */
	init_stat_params();

	/* Initialize process image client. */
	pclt_init();
	init_pservers();
	free_config();

	/* initialize connector and warn user if connector is not configured. */
	init_connector(connectorfile);
	
	/* Now initiate connections to all process image servers. */
	pclt_connect();
	
	/* detach from terminal and become daemon */
	if (!no_detach)
		daemonize();

	/* enter RT scheduler */
	realtime_enter(priority);
	
	/* drop privileges if we are root */
	drop_privileges(uid, gid);

	if (retry_timeout_msec > 0) {
		retry_tout = timeout_new(retry_timeout_handler, NULL);
		timeout_add(retry_tout, retry_timeout_msec);
	}

	/* Enable periodic timer. */
	if (cyclemsec > 0)
		periodic_timer_enable(cyclemsec);

	/* Run main cycle untill exit requested. */
	while (!terminate)
		run_cycle_once();

	realtime_exit();

	verbose("cycles_run %ld total_usec_last %ld total_usec_max %ld total_usec_min %ld", io_stats.cycles_run, io_stats.total_usec_last, io_stats.total_usec_max, io_stats.total_usec_min);
	verbose("cycle_usec %ld io_read_usec %ld io_write_usec %ld", io_stats.cycle_usec, io_stats.io_read_usec, io_stats.io_write_usec);

	/* Explicitely deinitialize everything. I want to be able to catch memoryleaks. */
	connector_deinit();

	pclt_deinit();

	par_tree_destroy();

	if (cyclemsec)
		periodic_timer_disable();

	if (io_wait) {
		io_del(io_wait, sigevent_fd());
		io_free(&io_wait);
	}

	if (retry_tout) {
		if (timeout_active(retry_tout))
			timeout_del(retry_tout);
		timeout_free(&retry_tout);
	}

	timeout_deinit();
	sigevent_deinit();

	return 0;
}
